;;;; A small library of common calendar definitions used in the CGI calendar interface stuff

;; The default calender column widht (in pixels)
(define default-column-width 180)

;; Minimum month width of a calendar
(define min-month-width 100)

;; Maximum month width of a calendar
(define max-month-width 800)

;; The directory where the calendar data (records and html files) are organized
(define data-directory "/user/normark/.public_html/cgi-bin/calendar/data/")

;; The URL of the place where the calendar cgi scripts reside.
(define calendar-url-prefix "http://www.cs.auc.dk/~normark/cgi-bin/calendar/")

;; The URL of the place where the generated calendars reside.
(define data-url-prefix (string-append calendar-url-prefix "data/"))


;; The maximum number of months in a calendar
(define max-calendar-length 24)  ; max 24 months, due to server efficiency

; (define seconds-in-a-day 86400)  from time library
(define seconds-in-a-half-day 43200)

;; The list of color. Each element is a string. Each string much correspond to
;; a symbol whose value is a rgb color list.
(define calendar-entry-color-list
  (list "light-blue" "blue" "light-green" "green" "light-red" "red" "light-purple" "purple" "light-grey" "grey" "light-yellow" "yellow"
        "light-brown" "brown" "light-orange" "orange" ))

;; Returns the rgb list of the color, as represented by the string color-string.
(define (string-to-color color-string)
  (cond ((equal? color-string "blue") (make-color 0 255 255))
        ((equal? color-string "light-blue") (make-color 185 255 255))
        ((equal? color-string "green") (make-color 15 255 158))
        ((equal? color-string "light-green") (make-color 141 255 209))
        ((equal? color-string "light-red") (make-color 249 149 149))
        ((equal? color-string "red") (make-color 255 0 0))
        ((equal? color-string "purple") (make-color 255 129 255))
        ((equal? color-string "light-purple") (make-color 255 201 255))
        ((equal? color-string "light-grey") (make-color 217 217 217))
        ((equal? color-string "grey") (make-color 178 178 178))
        ((equal? color-string "yellow") (make-color 255 255 0))
        ((equal? color-string "light-yellow") (make-color 255 255 157))
        ((equal? color-string "brown") (make-color 196 137 78))
        ((equal? color-string "light-brown") (make-color 216 177 138))
        ((equal? color-string "orange") (make-color 255 137 57))
        ((equal? color-string "light-orange") (make-color 255 185 137))
        (make-color 255 255 255)))


(define calendar-credit (con-space (con "The calendar software is designed and programmed by Kurt Nrmark, Department of Computer Science,
                                      Aalborg University, Denmark, using " (a-tag "http://www.cs.auc.dk/~normark/laml/" "LAML") " technology.") (br)
                                      (mail-link "normark@cs.auc.dk") ", "
                                      (a-tag "http://www.cs.auc.dk/~normark")))



; ---------------------------------------------------------------------------------------------------
;;; Calendar entry selectors.

;; Return the day of a calendar entry
(define cal-entry-day (make-selector-function 1))

;; Return the month of a calendar entry
(define cal-entry-month (make-selector-function 2))

;; Return the year of a calendar entry
(define cal-entry-year (make-selector-function 3))

;; Return whether we make afternoon start of a calendar entry
(define cal-entry-afternoon? (make-selector-function 4))

;; Return the number of halfdays of a calendar entry
(define cal-entry-halfdays (make-selector-function 5))

;; Return the period lenght of a calendar entry (7 is a week).
(define cal-entry-period-length (make-selector-function 6))

;; Return the repetition of a calendar entry.
(define cal-entry-period-multiplicity (make-selector-function 7))

;; Return the text of a calendar entry.
(define cal-entry-text (make-selector-function 8))

;; Return the color of a calendar entry.
(define cal-entry-color (make-selector-function 9))

;; Return the URL of a calendar entry.
(define cal-entry-url (make-selector-function 10))

;; Return from a calendar entry whether this entry should be deleted
(define cal-entry-delete? (make-selector-function 11))

;; Calendar entry constructor 
(define (make-calendar-entry day month year afternoon? halfdays period-lenght period-multiplicity text color url delete?)
  (list day month year afternoon? halfdays period-lenght period-multiplicity text color url delete?))


; ---------------------------------------------------------------------------------------------------
; Color key selectors:

(define color-key-text-of (make-selector-function 1 "color-key-text-of"))
(define color-key-color-of (make-selector-function 2 "color-key-color-of"))

; ---------------------------------------------------------------------------------------------------


;; Read calendar info (an alist) from file-name, normally of type string.
;; If the file-name is the symbol new, always return an empty calendar info (an empty list).
;; The file-name is taken from the directory data-directory (a free variable of this function).
(define (read-calendar-info file-name)
 (if (and (symbol? file-name) (eq? file-name 'new))
     '() 
     (let ((path (string-append data-directory file-name)))
       (if (file-exists? path) (file-read path) '()))))

;; Write the calender info (an alist) on file-name.
;; The file will be written in the directory data-directory (free variable of this function).
(define (write-calendar-info info file-name)
  (let ((path (string-append data-directory file-name)))
    (save-a-list info path)))

;; Return a fonted string which tells when this page is generated.
(define (when-generated)
 (let* ((dt (date-time (current-time)))
        (date (car dt))
        (time (cadr dt)))
  (font 2 red (con "Generated: " date ", " time))))


;; Make a calendar event in the format the original calendar tool and the calendar function expects.
;; The input, entry, to this function is an entry of the form:
;; (day month year number-of-half-days text color url).
;; All elements of the list are strings.
(define (make-calendar-event entry)
 (let* ((day (as-number (cal-entry-day entry)))
        (month (as-number (cal-entry-month entry)))
        (year (as-number (cal-entry-year entry)))
        (half-days (as-number (cal-entry-halfdays entry)))
        (afternoon? (as-boolean (cal-entry-afternoon? entry)))
        (end-displacement (* (- half-days 1) seconds-in-a-half-day))
       )
  (list 
   (if afternoon? (second-count year month day 12 0 1) (second-count year month day 0 0 1))
   (+ end-displacement (if afternoon? (second-count year month day 23 59 0) (second-count year month day 11 59 59)))
   (cal-entry-text entry)
   ""
   (string-to-color (cal-entry-color entry))
   (cal-entry-url entry)
  )
 )
)


(define (news-flash)
  (if (> (string-length news-flash-string) 0)
   (con (p) 
        (table-1 1 (list 600) (list yellow) 
          (list (list 
            (con (font 4 black  news-flash-string)))))
        (p))
    ""))



; ---------------------------------------------------------------------------------------------------
;;; ALIST TO LIST OF CALENDAR ENTRIES REORGANIZATION. 
;;; Functions which reorganize selected part of the input form-alist to
;;; a list of calendar entries. In the input alist, both keys and values are strings.
;;; The main function of this section: make-calendar-entries


;; Return a list of calendar entries given the input alist-input
;; extracted from the input forms of a browser.
;; There are no requirements on the mutual order of associations in alist-input.
(define (make-calendar-entries alist-input)
  (let* ((relevant-entry-alist (filter row-data? alist-input))
	 (which-rows (extract-row-numbers relevant-entry-alist))
         (alist-symbol-keys (map symbolize-key relevant-entry-alist))
	 (candidate-entries (reorganize-rows alist-symbol-keys which-rows)))
    (filter (negate empty-row?) candidate-entries)))  ; filter empty rows!

; Does association contain row data, i.e. table data from calendar entry
(define (row-data? association)
  (row-number (car association)))

; Return the row number (an integer) given a re-string.
; Rn-string is a string of the form "rn-xxx", where n is a one or more
; digit, positive number. Extract the number n. If rn-string is not of the appropriate
; form, return #f
(define (row-number rn-string)

 (if (eq? (string-ref rn-string 0) #\r)
     (let ((sp (split-point #\- rn-string)))
       (if sp
	   (as-number (substring rn-string 1 sp))
	   #f))
     #f))

(define (extract-row-numbers alist)
  (remove-duplicates
     (map (lambda (a) (row-number (car a))) alist)))

(define (reorganize-rows alist numbers)
  (if (null? numbers)
      '()
      (cons (make-one-reorganized-row alist (car numbers))
	    (reorganize-rows alist (cdr numbers)))))

(define (make-one-reorganized-row alist n)
  (let ((d (get (make-key "day" n) alist))
        (m (get (make-key "month" n) alist))
        (y (get (make-key "year" n) alist))
        (nhd (get (make-key "number-of-half-days" n) alist))
        (pl (get (make-key "period-length" n) alist))
        (pm (get (make-key "period-multiplicity" n) alist))
        (a (defaulted-get (make-key "afternoon" n) alist #f))
        (t (get (make-key "text" n) alist))
        (c (get (make-key "color" n) alist))
        (u (get (make-key "url" n) alist))
        (del (as-boolean (defaulted-get (make-key "delete" n) alist #f))))
    (make-calendar-entry d m y a nhd pl pm t c u del)))

(define (make-key str n)
  (as-symbol (string-append "r" (as-string n) "-" str)))

;; empty if the text position is empty
(define (empty-row? calendar-entry)
  (equal? "" (cal-entry-text calendar-entry)) ; perhaps also if only spaces
)

;; obsolete if the delete position is true
(define (entry-obsolete? calendar-entry)
  (cal-entry-delete? calendar-entry))

;; auto delete entry if the entry is entirely older than the current time
(define (auto-delete-entry? calendar-entry)
  (let ((entry-end-time (calculate-entry-end-time calendar-entry)))
     (< entry-end-time the-time)))

;; Return the entry end time (the time where this entry ends in time)
;; should, in a more advance version, take periods and repetitions into account
(define (calculate-entry-end-time entry)
 (let* ((day (as-number (cal-entry-day entry)))
        (month (as-number (cal-entry-month entry)))
        (year (as-number (cal-entry-year entry)))
        (half-days (as-number (cal-entry-halfdays entry)))
        (period-length (as-number (cal-entry-period-length entry)))
        (period-repetition (as-number (cal-entry-period-multiplicity entry)))
        (afternoon? (as-boolean (cal-entry-afternoon? entry)))
        (base-time (if afternoon? (second-count year month day 23 59 0) (second-count year month day 11 59 59))))
    (+ base-time 
       (* (- half-days 1) seconds-in-a-half-day) 
       (* period-length seconds-in-a-day (- period-repetition 1)))
 ))

;; Return the start time a calendar entry
(define (calculate-entry-start-time entry)
 (let* ((day (as-number (cal-entry-day entry)))
        (month (as-number (cal-entry-month entry)))
        (year (as-number (cal-entry-year entry)))
        (afternoon? (as-boolean (cal-entry-afternoon? entry)))
        (base-time (if afternoon? (second-count year month day 12 0 1) (second-count year month day 0 0 1))))
   base-time
 )
)

;;; return the latest time of the entries in entry-list
(define (entry-list-end-time entry-list)
  (apply max
         (map calculate-entry-end-time entry-list)))

;;; return the earliest time of the entries in entry-list
(define (entry-list-start-time entry-list)
  (apply min
         (map calculate-entry-end-time entry-list)))

; ---------------------------------------------------------------------------------------------------
; Calendar entry list expanding.
; Repetitive entry are all exanded to 'one repetition' entries.
; Hereby the list becomes longer, but simpler.

(define (expand-calendar-repetitions calendar-entry-list)
  (expand-calendar-repetitions-1 calendar-entry-list '() ))

(define (expand-calendar-repetitions-1 calendar-entry-list res-entry-list)
  (cond ((null? calendar-entry-list) res-entry-list)
        ((single-repetition-entry? (car calendar-entry-list))
           (expand-calendar-repetitions-1 
              (cdr calendar-entry-list)
              (cons (car calendar-entry-list) res-entry-list)
           ))
        ((multi-repetition-entry? (car calendar-entry-list)) 
           (expand-calendar-repetitions-1 
              (cons (repetition-reduce (car calendar-entry-list)) (cdr calendar-entry-list))
              (cons (force-one-repetition (car calendar-entry-list)) res-entry-list)))
        (else (error "expand-calendar-repetitions-1: This case should not happen"))))

(define (single-repetition-entry? entry)
  (= (as-number (cal-entry-period-multiplicity entry)) 1))

(define (multi-repetition-entry? entry)
  (> (as-number (cal-entry-period-multiplicity entry)) 1))

(define (force-one-repetition entry)
  (let ((day (cal-entry-day entry))
        (month (cal-entry-month entry))
        (year (cal-entry-year entry))
        (afternoon (cal-entry-afternoon? entry))
        (halfdays (cal-entry-halfdays entry))
        (per-length (cal-entry-period-length entry))
        (repetition (cal-entry-period-multiplicity entry))
        (text (cal-entry-text entry))
        (color (cal-entry-color entry))
        (url (cal-entry-url entry))
        (del (cal-entry-delete? entry)))
    (let* ((rep-date (make-period-date (as-number year) (as-number month) (as-number day) (- (as-number repetition) 1) (as-number per-length)))
           (rep-year (car rep-date))
           (rep-month (cadr rep-date))
           (rep-day (caddr rep-date)))
   (make-calendar-entry (as-string rep-day) (as-string rep-month) (as-string rep-year) afternoon halfdays per-length (as-string 1)
                        (string-append text " " repetition) color url del))))

;; given a date in terms of year, month, and day (all integers)
;; an an integer rep
;; return a list of (year month day) which is rep times period-length into the future
(define (make-period-date year month day rep period-length)
  (let* ((input-second-count (second-count year month day 0 0 0))
         (future-second-count (+ input-second-count (* rep period-length seconds-in-a-day)))
         (future-time-record (time-decode future-second-count)))
   (list (car future-time-record) (cadr future-time-record) (caddr future-time-record))))
   

(define (repetition-reduce entry)
  (let ((day (cal-entry-day entry))
        (month (cal-entry-month entry))
        (year (cal-entry-year entry))
        (afternoon (cal-entry-afternoon? entry))
        (halfdays (cal-entry-halfdays entry))
        (per-length (cal-entry-period-length entry))
        (repetition (as-number (cal-entry-period-multiplicity entry)))
        (text (cal-entry-text entry))
        (color (cal-entry-color entry))
        (url (cal-entry-url entry))
        (del (cal-entry-delete? entry)))
   (make-calendar-entry day month year afternoon halfdays per-length (as-string (- repetition 1)) text color url del)))


; ---------------------------------------------------------------------------------------------------
; REMAINING FUNCTIONALITY:

;; Return a data structure (an alist) which represents the calendar on a file.
;; The input-alist is an alist, in which the keys have to be symbols.
(define (make-persistent-calendar-data-structure entry-list color-key-list input-alist)
  (alist-from-keys-and-values
    (list "owner" "password" "description" "start-month" "start-year" "number-of-months" "column-width" "auto-delete" 
          "font-size" "language" "week-separator"  "merge-1" "merge-2" "merge-3" "merge-4" "merge-5" "merge-6" "merge-7"  "entries" "color-keys" "time"
          "purpose" "maintainer"  
          "auto-update-nine-am" "auto-update-noon" "auto-update-three-pm" "auto-update-midnight"
    )
    (list (get 'owner input-alist)
          (crypt-string (get 'password input-alist))
          (get 'description input-alist)
          (get 'start-month input-alist)
          (get 'start-year input-alist)
          (get 'number-of-months input-alist)
          (get 'column-width input-alist)
          (as-boolean (defaulted-get 'auto-delete input-alist #f))
          (get 'font-size input-alist)
          (get 'language input-alist)
          (get 'week-separator input-alist)

          (defaulted-get 'merge-1 input-alist "")
          (defaulted-get 'merge-2 input-alist "")
          (defaulted-get 'merge-3 input-alist "")
          (defaulted-get 'merge-4 input-alist "")
          (defaulted-get 'merge-5 input-alist "")
          (defaulted-get 'merge-6 input-alist "")
          (defaulted-get 'merge-7 input-alist "")

          entry-list
          color-key-list
          the-time
          purpose
          maintainer
          auto-update-nine-am auto-update-noon auto-update-three-pm auto-update-midnight
    )))

; ----------------------------------------------------------------------------------------------------

; -----------------------------------------------------------------------------
; EXTRACTION AND ORGANIZATION OF COLOR KEY FIELDS.
; Here we extract the ci-key-text and ci-key-color fields from the form input.
; Notice that we assume a fixed format a naming of the relevant fields.
; Not very good programming.

(define (extract-color-key-list form-a-list-symbols)
  (let ((color-key-fields  ; the list of ci-key-text names, not ci-key-color names. Symbols
         (filter 
          (lambda (key) (color-key? key))
          (map car form-a-list-symbols))
        )
       )
   (filter 
    (lambda (color-key)
      (not (empty-string? (color-key-text-of color-key))))
    (map 
     (lambda (color-key-text-name)
       (let ((similar-color-field-name (color-key-color-name-given-text color-key-text-name)))
         (make-color-key-entry 
           (get color-key-text-name form-a-list-symbols)
           (get similar-color-field-name form-a-list-symbols))))
     color-key-fields))))

(define (make-color-key-entry text color)
  (list text color))

; If the key a color-key symbol, such as 'c1-key-text
(define (color-key? key-symbol)
 (let ((key-string (as-string key-symbol)))
   (and (>= (string-length key-string) 11)
        (equal? (substring key-string 0 1) "c")
        (equal? (substring key-string 3 (string-length key-string)) "key-text"))))

; Given ci-key-text, return ci-key-color
; As a pre-condition (color-key? (as-symbol text-name))
(define (color-key-color-name-given-text text-name)
 (let ((text-name-as-string (as-string text-name)))
  (as-symbol
    (string-append (substring text-name-as-string 0 6) "-color"))))


; -----------------------------------------------------------------------------
; COLOR KEY PRESENATION:

(define (color-key-legend color-key-list)
  (table-3
    0
    (make-list (length color-key-list) 80)
    (list
      (map color-key-legend-entry color-key-list))))

(define (color-key-legend color-key-list)
  (if (not (null? color-key-list))
      (multi-column-list
         6
         (map color-key-legend-entry color-key-list)
         800)
      ""))

(define (color-key-legend-entry entry)
  (let ((text (color-key-text-of entry))
        (color (color-key-color-of entry)))
    (color-frame (font-size 2 text) (string-to-color color))))

; ---------------------------------------------------------------------------------------------------

; ---------------------------------------------------------------------------------------------------
; We define a function which finds out whether a merge-entries is original or on interactive format
; The funtion returns the entries in original format:

; Entries is either a list structure, as saved by make-persistent-calendar-data-structure,
; or a list structure of the form used for the calendar function's calendar-events list.
; Return, in all cases, a calendar-events list format.
; Does NO recursive merging - be careful with cyclic merging!
(define (normalize-to-calendar-event-format entries)
  (if (interactive-calendar-entries? entries) 
      (let ((sym-entries (map symbolize-key entries))) 
        (map make-calendar-event (expand-calendar-repetitions (defaulted-get 'entries sym-entries '()))))
      entries))

(define (interactive-calendar-entries? entries)
  (if (not (null? entries))
      (let ((first-entry (car entries)))
        (or (symbol? (car first-entry)) (string? (car first-entry))))
      #f))

; ---------------------------------------------------------------------------------------------------

;; Mutate a-list with new-pair. We assume that keys are symbols, due
;; to comparison with eq?. Return the possibly mutated a-list.
(define (mutate-a-list! new-pair a-list)
 (let* ((key (car new-pair))
        (val (cdr new-pair))
        (existing-pair (find-in-list (lambda (x) (eq? (car x) key)) a-list))
       )
    (if existing-pair 
        (set-cdr! existing-pair val))
    a-list))


; ---------------------------------------------------------------------------------------------------

; add owner to all-auto.dat if he/she is not alread added
(define (add-auto-update owner)
 (let* ((auto-update-file-name (string-append data-directory "all-auto.dat"))
        (auto-update-list (file-read auto-update-file-name))
      )
   (if (not (member owner auto-update-list))
       (file-write 
         (cons owner auto-update-list)
         auto-update-file-name))))

; withdraw owner from all-auto.dat, if he/she is on the list
(define (withdraw-auto-update owner)
 (let* ((auto-update-file-name (string-append data-directory "all-auto.dat"))
        (auto-update-list (file-read auto-update-file-name))
      )
   (if (member owner auto-update-list)
       (file-write 
         (filter (lambda (x) (not (equal? x owner))) auto-update-list)
         auto-update-file-name))))

; Compare request of auto update times with the actual time
(define (actual-time-to-update-this-calendar?)
  (or 
    (and auto-update-nine-am (is-time-around? 9))
    (and auto-update-noon (is-time-around? 12))
    (and auto-update-three-pm (is-time-around? 15))
    (and auto-update-midnight (is-time-around? 0))))

; Is the actual time in the neighborhood of n (n is an integer between 0 and 24).
(define (is-time-around? n)
  (let* ((tm (current-time))
         (tm-str (time-decode tm))
         (hr (fourth tm-str))
        )
    (cond ((= n 0) (or (= hr 23) (= hr 24) (= hr 0) (= hr 1) ))
          (else (or (= hr n) (= hr (- n 1)) (= hr (+ n 1)))))))
 
  

